/******************************************************************************
*
* LASER POSITION SENSOR
*
* Filename:		main.c
* Author:		  Joe Grand, Grand Idea Studio, Inc. [www.grandideastudio.com]
*
* Last Updated: November 10, 2009
*
* Description:	Main File (Freescale MC9S08JM8)
*
* Distributed under a Creative Commons Attribution-Share Alike 3.0 Unported license:
* http://creativecommons.org/licenses/by-sa/3.0/
*		
*******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "LPS.h"
#include "serial.h"


/****************************************************************************
 ************************** Global variables ********************************
 ***************************************************************************/

struct ad_data newdata; // used in TPM1CH0_ISR

byte LED_dir;           // for red LED dimming control
word LED_dutycycle;

byte ready_flag;


/****************************************************************************
 ************************** Function definitions ****************************
 ***************************************************************************/
void main(void)
{
  HW_Init();
  
  ready_flag = 0; 
  LEDG = ON; // both LEDs on for power-on test and to visually show system is getting ready
  TPM2C0V = PWM_DUTYCYCLE; 
        
  coefs.a = COEF_A;	// init the iir filter coefficients  
	coefs.b = COEF_B;
	
	getvoltageoffset();   // zero the voltage offsets from the detector and roomlights (get our null values)
  
  while(1)
  {    
    if (!SW_EN) // if enable is pulled low, the host wants to enable the LRF
    {
      delay_ms(25); // poor man's debounce
      
      while (!SW_EN) // if the enable line is still low...
      {
        if (ready_flag == 0) 
        {
            SendMsg("*"); // send flag one time after /ENABLE is pulled low and system is ready to go
            ready_flag = 1;
        }
      
        LEDG = OFF;
        RTCSC_RTIE = 1; // enable RTC to allow PWM dimming of red LED 
  
        compute_xy();	// compute the new xy (calibrated) data...  
			  display_xy();	// ...and send it out the serial port
      }  
    }
    
    // enter idle state
    LEDG = ON;
    RTCSC_RTIE = 0; // red LED off 
    TPM2C0V = 0;
    ready_flag = 0;    
    delay_ms(10);   // settle
    
    // go to sleep until the /ENABLE line is pulled LOW
    SPMSC1_LVDE = 0; // disable low-voltage detect
    SPMSC1_LVDSE = 0;
    SPMSC2_PPDC = 0; // partial power down control: stop3 mode  
    asm ( stop; ); // enter stop3 mode, will wake on /RESET or interrupt
  }
}


/********************************************************/
// iir_filter_data(olddata, newdata)
//
//	IIR filtering
// 	
// 	Y(i) = A*Y(i-1) + B*(X(i) + X(i-1))
//	A = (2/ts - 2PI*f) / (2/ts + 2PI*f)
//	B = 2PI*f / (2/ts + 2PI*f)
//
signed int iir_filter_data(signed int olddata, signed int newdata, signed int lastsample)
{
	signed long filter;

	filter = ((signed long) olddata * coefs.a) + (((signed long) newdata + (signed long) lastsample) * coefs.b);
	olddata = (signed int) (filter >> 15);
	return (olddata);
}


/********************************************************/
// compute_xy()
//
//	take the filtered raw data readings and compute calibrated X, Y data
//  subtract the display offsets from the X, Y data.
//
//  -L/2 < x or y < +Y/2 
// 
//	x = ((x2 + y1) - (x1 + y2)) / (x1 + x2 + y1 + y2)
//	y = ((x2 + y2) - (x1 + y1)) / (x1 + x2 + y1 + y2)
//	Power = (x1 + x2 + y1 + y2) / powercoeff
//
void compute_xy(void)
{
	signed long diff, sum;
	struct ad_data compute;
	
	// make local copies of the raw data
	TPM1C0SC_CH0IE = 0;	// disable timer1 interrupt, make sure the readings don't change while doing the calculations
	compute.x_1 = rawdata.x_1;
	compute.x_2 = rawdata.x_2;
	compute.y_1 = rawdata.y_1;
	compute.y_2 = rawdata.y_2;
	compute.error = rawdata.error;
  TPM1C0SC_CH0IE = 1; // re-enable interrupt
  
	// compute the sum
	sum = compute.x_1 + compute.x_2 + compute.y_1 + compute.y_2;
	
	// do x
	diff = (compute.x_2 + compute.y_1) - (compute.x_1 + compute.y_2);	
	displaydata.x = (signed int) ((diff * DET_LENGTH) / sum) - d_offsets.x;
	
	// do y
	diff = (compute.x_2 + compute.y_2) - (compute.x_1 + compute.y_1);
	displaydata.y = (signed int) ((diff * DET_LENGTH) / sum) - d_offsets.y;	
}


/********************************************************/
//	display_xy()
//
//	convert the calculated x and y values to ascii
//  and output them via serial interface
//
void display_xy(void)
{
	unsigned char s[8];
	signed int answer;

  if (rawdata.error == NOERROR)
  {
  	answer = displaydata.x;
  	s[0] = (answer < 0) ? '-' : '+';
  	answer = abs(answer);
  	s[1] = '0' + (unsigned char) (answer / 1000);
  	answer = answer % 1000;
  	s[2] = '0' + (unsigned char) (answer / 100);
  	answer = answer % 100;
  	s[3] = '0' + (unsigned char) (answer / 10);
  	answer = answer % 10;
  	s[4] = '0' + (unsigned char) answer;
  	s[5] = '\0';

    SendMsg("X");
    SendMsg(s);
    SendMsg(" ");
    
   	answer = displaydata.y;
  	s[0] = (answer < 0) ? '-' : '+';
  	answer = abs(answer);
  	s[1] = '0' + (unsigned char) (answer / 1000);
  	answer = answer % 1000;
  	s[2] = '0' + (unsigned char) (answer / 100);
  	answer = answer % 100;
  	s[3] = '0' + (unsigned char) (answer / 10);
  	answer = answer % 10;
  	s[4] = '0' + (unsigned char) answer;
  	s[5] = '\0';

    SendMsg("Y");
    SendMsg(s);
    SendMsg("\n\r");
  }
}


/********************************************************/
//	getvoltageoffset()
//
//	get the current raw filtered a/d values and use them as an offset to be subracted from subsequent
//	a/d readings
//	this is important because these offsets will affect the calculations of x and y, especially when
//	signal levels are small
//
void getvoltageoffset(void)
{
	unsigned char i;
	
	// important to zero any old offset values we might have
	v_offsets.x_1 = 0;
	v_offsets.x_2 = 0;
	v_offsets.y_1 = 0;
	v_offsets.y_2 = 0;

	// allow the filtered data to accumulate and put out some '.' to let user know it's doing something
	for(i = 0; i < 16; i++)
	{
		delay_ms(250);  // the interrupt routine is always bringing in new data and filtering it
	}

	// make the new offsets equal to the current filtered a/d readings
	TPM1C0SC_CH0IE = 0;	// disable timer1 interrupt, make sure the readings don't change while doing the calculations
	v_offsets.x_1 = rawdata.x_1;
	v_offsets.x_2 = rawdata.x_2;
	v_offsets.y_1 = rawdata.y_1;
	v_offsets.y_2 = rawdata.y_2;
  TPM1C0SC_CH0IE = 1; // re-enable interrupt
}


/********************************************************/
void HW_Init(void)
{ 

  DisableInterrupts;
  delay_ms(100);

  // copy trim values stored in Flash in order to guarantee an accurate clock
  asm
  {
    lda $FFAF
    sta MCGTRM   ; Load TRIM bits from Flash and store it into MCGTRM

    lda MCGSC
    and #$FE
    ora $FFAE   ; Load FTRIM bit from Flash and store it in the proper field of MCGSC
    sta MCGSC
  }    
     
  /* Common initialization of the write once registers */
  /* SOPT1: COPT=0,STOPE=1,??=1,??=0,??=0,??=1,??=1 */
  // disable COP/watchdog
  SOPT1 = 0x33;
  
  /* SPMSC1: LVWF=0,LVWACK=0,LVWIE=0,LVDRE=1,LVDSE=1,LVDE=1,??=0,BGBE=0 */
  SPMSC1 = 0x1C; 
  
  /* SPMSC2: ??=0,??=0,LVDV=0,LVWV=0,PPDF=0,PPDACK=0,??=0,PPDC=0 */
  SPMSC2 = 0x00;
  
  /* System clock initialization */
  // device defaults to FLL Engaged Internal (FEI) mode upon power-on
  
  /* MCGC1: CLKS=0,RDIV=0,IREFS=1,IRCLKEN=1,IREFSTEN=0 */
  MCGC1 = 0x06;

  /* MCGC2: BDIV=0,RANGE=0,HGO=0,LP=0,EREFS=0,ERCLKEN=0,EREFSTEN=0 */
  // divide-by-1 for bus frequency of 16MHz
  MCGC2 = 0x00;
  
  /* MCGC3: LOLIE=0,PLLS=0,CME=0,??=0,VDIV=1 */
  // VDIV = 1 = reference clock * 4
  MCGC3 = 0x01; 
  
  while(MCGSC_OSCINIT) {               /* Wait until OSC is initialized (MCG mode = FEI) */
  }
  while(!MCGSC_IREFST) {               /* Wait until internal reference is selected */
  }
  while(!MCGSC_LOCK) {                 /* Wait until FLL is locked */
  }
  while((MCGSC & 0x0C) != 0x00) {      /* Wait until FLL clock is selected as a bus clock reference */
  }
       
  // After reset, all pins are configured as high-impedance general-purpose
  // inputs with internal pullup devices disabled
  
    /*** LED (Green) ***/
  LEDG_DD = LEDR_DD = _OUT; // set LED pins to outputs
  LEDG_DS = LEDR_DS = _OUT; // set LED pins to high output drive strength

  /*** /ENABLE ***/  
  KBISC_KBIE = 0;   // disable interrupts for KBI module 
  KBIES_KBEDG6 = 0; // falling edge
  PTGPE_PTGPE2 = 1; // enable internal pullup
  KBIPE_KBIPE6 = 1; // enable KBI interrupt on pin
  KBISC_KBACK = 1;  // clear any false interrupts
  KBISC_KBIE = 1;   // enable interrupts for KBI module

  // Set unused/unconnected pins to outputs so the pins don't float
  // and cause extra current drain
  // data register default = LOW
  PTCDD_PTCDD0 = PTCDD_PTCDD1 = _OUT;
  PTDDD_PTDDD2 = _OUT;
  PTEDD_PTEDD2 = PTEDD_PTEDD3 = PTEDD_PTEDD4 = PTEDD_PTEDD5 = PTEDD_PTEDD6 = PTEDD_PTEDD7 = _OUT;
  PTGDD_PTGDD3 = PTGDD_PTGDD4 = PTGDD_PTGDD5 = _OUT;

  /*** PWM (Timer 2), Channel 0: LED (Red) ***/
  TPM2SC_TOIE = 0;       // disable timer overflow interrupt
  TPM2SC_CPWMS = 0;      // each channel function set by MSnB:MSnA control bits
  TPM2SC_CLKSx = 0b01;   // clock select = bus
  TPM2SC_PS = 0b000;     // prescaler = /1
  
  //LEDR = LOW;
  TPM2C0SC_CH0IE = 0;    // disable channel 1 interrupt
  TPM2C0SC_MS0x = 0b10;  // edge-aligned PWM
  TPM2C0SC_ELS0x = 0b10; // high-true pulses
  TPM2MOD = PWM_PERIOD;    // PWM period
  TPM2C0V = PWM_DUTYCYCLE; // PWM duty cycle
  LED_dir = LED_dutycycle = 0;
 
  /*** Timer 1, periodic A/D measurements of PSD ***/
  TPM1SC_TOIE = 0;       // disable timer overflow interrupt
  TPM1SC_CPWMS = 0;      // each channel function set by MSnB:MSnA control bits
  TPM1SC_CLKSx = 0b01;   // clock select = bus
  TPM1SC_PS = 0b010;     // prescaler = /4  
  
  TPM1C0SC_CH0IE = 1;       // enable channel 0 interrupt
  TPM1C0SC_MS0x = 0b01;     // output compare
  TPM1C0SC_ELS0x = 0b01;    // toggle output on compare  
  TPM1MOD = T1_SAMPLERATE;  // set timer overflow period 

  /*** RTC ***/
  RTCMOD = 0;
  RTCSC_RTIE = 0;       // disable interrupt module (enable when we want to pulse red LED)

  RTCSC_RTCLKS = 0b10;  // clock source = internal reference clock (31.25kHz)
  RTCSC_RTCPS = 0b1000; // prescaler period = 31.25uS
         
  /*** ADC ***/
  ADCCFG_MODE = 0b01;   // 12-bit conversion
  APCTL1_ADPC4 = 1;     // set pins for A/D control, all other pins are I/O
  APCTL1_ADPC5 = 1;
  APCTL2_ADPC8 = 1;
  APCTL2_ADPC9 = 1;
  
    /*** SCI/UART ***/
  // Loop mode = Disabled (Normal operation)
  // TX and RX = Enabled
  // Data format = 8 bits
  // Stop in Wait mode = Disabled
  // Wake up = Idle line wakeup
  // Idle character counting = After start bit
  // Parity = None
  // Send break = Disabled
  // Receiver wakeup = Normal operation
  // Break character generation length = 10 or 11 bits
  // TX output = Not inverted 
  SCI1C1 = 0x00;
  SCI1C2 = 0x0C;
  SCI1C3 = 0x00;
  SCI1D = 0x00;
         
  // SCI baud rate = BUSCLK/(16 * BR)
  SCI1BDH = 0x00;
  SCI1BDL = 0x68; // 9600 bps (9615.385 baud @ 16MHz bus clock)  
  //SCI1BDL = 0x34; // 19.2 kbps
  //SCI1BDL = 0x1A; // 38.4 kbps

  EnableInterrupts;
}
 

/**************************************************************
*	Function: 	delay
*	Parameters: 16-bit delay value
*	Return:		none
*
*	Simple delay loops
* These are not very accurate because they could get interrupted
*
**************************************************************/
 
void delay_ms(unsigned int ms)  // delay the specified number of milliseconds
{
  unsigned int timer;

  while (ms != 0) 
  {
    // this number is dependent on the bus clock frequency (currently at 16MHz)
    for (timer=0; timer <= 752; timer++); 
    ms--;
  }
}

void delay(unsigned int count) 
{
	unsigned int i;
	for (i = 0; i < count; i++);
}

/*******************************************************
********************************************************
** Interrupt handlers                                 **
********************************************************
*******************************************************/

interrupt VectorNumber_Vkeyboard void KBI_ISR(void) 
{
  KBISC_KBACK = 1;
}


interrupt VectorNumber_Vrtc void RTC_ISR(void)
{
  RTCSC_RTIF = 1;  // clear interrupt flag
  
  if (LED_dir == 0) // count up
  {
    LED_dutycycle++;
    if (LED_dutycycle >= PWM_DUTYCYCLE) LED_dir = 1;  // if we've reached our maximum desired duty cycle, reverse direction
  }
  else // count down
  {
    LED_dutycycle--;
    if (LED_dutycycle == 0) LED_dir = 0;
  }
  
  TPM2C0V = LED_dutycycle;  // set PWM duty cycle for this period
}

  
interrupt VectorNumber_Vtpm1ch0 void TPM1CH0_ISR(void) 
{
  unsigned char tmp;
  
  tmp = TPM1C0SC;    // read register per Table 16-5, field 7   
  TPM1C0SC_CH0F = 0; // clear interrupt flag

  //	get new a/d readings for PSD's X1, X2, Y1, Y2
  //	bounds check the new readings
  //	put them into the IIR filter
  ADCSC1 = X1_INPUT;
  while(!ADCSC1_COCO);
  newdata.x_1 = ADCR;
        
  delay(10);

  ADCSC1 = X2_INPUT; 
  while(!ADCSC1_COCO);
  newdata.x_2 = ADCR;

  delay(10);
                  
  ADCSC1 = Y1_INPUT;
  while(!ADCSC1_COCO);
  newdata.y_1 = ADCR;

  delay(10);
        
  ADCSC1 = Y2_INPUT;
  while(!ADCSC1_COCO);
  newdata.y_2 = ADCR;

	// check for a saturated condition on the a/d converter
	if ((newdata.x_1 > SATLEVEL) || (newdata.x_2 > SATLEVEL) ||
		  (newdata.y_1 > SATLEVEL) || (newdata.y_2 > SATLEVEL) )
	{
		rawdata.error = DETSAT;
	}
	else
		rawdata.error = NOERROR;
	
	// subtract out the offset voltages from the raw a/d readings
	newdata.x_1 = newdata.x_1 - v_offsets.x_1;
	newdata.x_2 = newdata.x_2 - v_offsets.x_2;
	newdata.y_1 = newdata.y_1 - v_offsets.y_1;
	newdata.y_2 = newdata.y_2 - v_offsets.y_2;

	// if < 0 make them 0,  negative voltages confuse our equations so in our
	// system there are no negative voltages
	newdata.x_1 = (newdata.x_1 < 0) ? 0 : newdata.x_1;
	newdata.x_2 = (newdata.x_2 < 0) ? 0 : newdata.x_2;
	newdata.y_1 = (newdata.y_1 < 0) ? 0 : newdata.y_1;
	newdata.y_2 = (newdata.y_2 < 0) ? 0 : newdata.y_2;

	// check for a sum too low condition
	if ((newdata.x_1 + newdata.x_2 + newdata.y_1 + newdata.y_2) < LOWSUMLEVEL)
	{
		rawdata.error = LOWSUM;
	}	
	
	// IIR filtering
	if (!rawdata.error)
	{
		rawdata.x_1 = iir_filter_data(rawdata.x_1, newdata.x_1, last_ad_sample.x_1);
		rawdata.x_2 = iir_filter_data(rawdata.x_2, newdata.x_2, last_ad_sample.x_2);
		rawdata.y_1 = iir_filter_data(rawdata.y_1, newdata.y_1, last_ad_sample.y_1);
		rawdata.y_2 = iir_filter_data(rawdata.y_2, newdata.y_2, last_ad_sample.y_2);
	}

	// load these values for the next samples iir filter
	last_ad_sample.x_1 = newdata.x_1;
	last_ad_sample.x_2 = newdata.x_2;
	last_ad_sample.y_1 = newdata.y_1;
	last_ad_sample.y_2 = newdata.y_2;
}

    
/****************************** END OF FILE **********************************/

